@@ -11,6 +11,7 @@
  * 2003-jan: Support in ntreg for adding keys etc. Editor updated.
  * 2003-jan: Changed to use more of struct based V + some small stuff
  * 2004-jan: Changed some of the verbose/debug stuff
+ * 2004-aug: More stuff in regedit. Stringinput bugfixes.
  * 
  * Copyright (c) 1997-2004 Petter Nordahl-Hagen.
  * Freely distributable in source or binary for noncommercial purposes,
@@ -57,7 +58,7 @@
 #include "ntreg.h"
 #include "sam.h"
 
-const char chntpw_version[] = "chntpw version 0.99.2 040105, (c) Petter N Hagen";
+const char chntpw_version[] = "chntpw version 0.99.3 040818, (c) Petter N Hagen";
 
 extern char *val_types[REG_MAX+1];
 
@@ -232,7 +233,6 @@ short handle_F(int rid, int mode)
   unsigned short acb;
   int b;
 
-
   if (H_SAM < 0) return(0);
 
   /* Get users F value */
@@ -661,7 +661,7 @@ char *change_pw(char *buf, int rid, int vlen, int stat)
  * MULTI_SZ (multi unicode-string) - only displays first string,
  * but also does a hexdump.
  */
-void cat_vk(struct hive *hdesc, int nkofs, char *path)
+void cat_vk(struct hive *hdesc, int nkofs, char *path, int dohex)
 {     
   void *data;
   int len,i,type;
@@ -689,6 +689,7 @@ void cat_vk(struct hive *hdesc, int nkofs, char *path)
   printf("Value <%s> of type %s, data length %d [0x%x]\n", path,
 	 (type < REG_MAX ? val_types[type] : "(unknown)"), len, len);
 
+  if (dohex) type = REG_BINARY;
   switch (type) {
   case REG_SZ:
   case REG_EXPAND_SZ:
@@ -737,6 +738,9 @@ struct cmds {
 #define MCMD_DELVALL 15
 #define MCMD_NEWKEY 16
 #define MCMD_DELKEY 17
+#define MCMD_CATHEX 18
+#define MCMD_RDEL 19
+#define MCMD_CK 20
 
 struct cmds maincmds[] = {
  { "cd" , MCMD_CD } ,
@@ -760,6 +764,9 @@ struct cmds maincmds[] = {
  { "delallv", MCMD_DELVALL } ,
  { "nk", MCMD_NEWKEY } ,
  { "dk", MCMD_DELKEY } ,
+ { "hex", MCMD_CATHEX } ,
+ { "rdel", MCMD_RDEL } ,
+ { "ck", MCMD_CK } ,
  { "?", MCMD_HELP } ,
  { "", 0 }
 };
@@ -774,7 +781,7 @@ void edit_val(struct hive *h, int nkofs, char *path)
   int type,len,n,i,in,go, newsize, d = 0, done, insert = 0;
   char inbuf[SZ_MAX+4];
   char origstring[SZ_MAX+4];
-  char *newstring = NULL;
+  char *newstring;
   char *dbuf;
 
   type = get_val_type(h, nkofs, path);
@@ -808,6 +815,7 @@ void edit_val(struct hive *h, int nkofs, char *path)
   case REG_SZ:
   case REG_EXPAND_SZ:
   case REG_MULTI_SZ:
+    newstring = NULL;
     dbuf = (char *)&kv->data;
     cheap_uni2ascii(dbuf,origstring,len);
     n = 0; i = 0;
@@ -828,30 +836,41 @@ void edit_val(struct hive *h, int nkofs, char *path)
     }
     n = 0; i = 0; in = 0; go = 0; done = 0;
 
+    /* Now this one is RATHER UGLY :-} */
+
     while (i < (len>>1)-1 || !done) {
+
       printf("[%2d]: %s\n",n, insert == 1 ? "[INSERT]" : ((i < (len>>1)-1 ) ? origstring+i : "[NEW]"));
       if (insert) insert++;
       if (!go) fmyinput("-> ",inbuf, 500);
       else *inbuf = 0;
       if (*inbuf && strcmp("--q", inbuf)) {
 	if (!strcmp("--n", inbuf) || !strcmp("--Q", inbuf)) { /* Zap rest */
-	  i = (len>>1) +1; done = 1;
+	  i = (len>>1) ; done = 1;
 	} else if (strcmp("--i", inbuf)) {  /* Copy out given string */
 	  if (!strcmp("--e",inbuf)) *inbuf = '\0';
 	  if (newstring) newstring = realloc(newstring, in+strlen(inbuf)+1);
 	  else newstring = malloc(in+strlen(inbuf)+1);
-	  strncpy(newstring+in, inbuf, 499);
+	  strcpy(newstring+in, inbuf);
 	  in += strlen(inbuf)+1;
 	} else {
 	  insert = 1;
 	}
       } else {  /* Copy out default string */
+
 	if (newstring) newstring = realloc(newstring, in+strlen(origstring+i)+1);
 	else newstring = malloc(in + strlen(origstring+i) + 1);
-	strncpy(newstring+in, origstring+i, 499);
+	strcpy(newstring+in, origstring+i);
 	in += strlen(origstring+i)+1;
-	if (!strcmp("--q", inbuf)) { go = 1; done = 1; }
+
+	if (!strcmp("--q", inbuf)) { 
+	  go = 1; done = 1;
+	  if (!(i < (len>>1)-1 )) {
+	    in--;  /* remove last empty if in NEW-mode */
+	  }
+	}
       }
+      
       if (!insert) i += strlen(origstring+i) + 1;
       if (insert != 1) n++;
       if (insert == 2) insert = 0;
@@ -865,17 +884,21 @@ void edit_val(struct hive *h, int nkofs, char *path)
     if (strcmp("--Q", inbuf)) {  /* We didn't bail out */
       if (newstring) newstring = realloc(newstring, in+1);
       else newstring = malloc(in+1);
-      in++;
-      *(newstring+in) = '\0';  /* Must add null termination */
-
+      if (type == REG_MULTI_SZ) {
+	in++;
+	*(newstring+in) = '\0';  /* Must add null termination */
+      }
       ALLOC(newkv,1,(in<<1)+sizeof(int));
       newkv->len = in<<1;
+      printf("newkv->len: %d\n",newkv->len);
       cheap_ascii2uni(newstring, (char *)&(newkv->data), in);
       
       d = 1;
+
+      FREE(kv);
+      kv = newkv;
+
     }
-    FREE(kv);
-    kv = newkv;
     break;
 
   default:
@@ -936,6 +959,7 @@ void mainloop(void)
   char path[1000];
   int l, vkofs, nh;
   int usehive = 0;
+  struct keyval *kv;
 
 #if ALLOC_DEBUG
   int pagestart;
@@ -970,14 +994,16 @@ void mainloop(void)
 	printf("Simple registry editor:\n");
 	printf("hive [<n>] - list loaded hives or switch to hive numer n'\n");
 	printf("cd <key> - change key\nls | dir [<key>] - show subkeys & values,\n");
-        printf("cat | type <value> - show key value\nst [<hexaddr>] - show struct info\n");
+        printf("cat | type <value> - show key value\nhex <value> - hexdump of value data\nst [<hexaddr>] - show struct info\n");
   /* printf("pw | passwd [<hexaddr>] - try the password routine on struct at <hexaddr>\n"); */
+	printf("ck [<keyname>]          - Show keys class data, if it has any\n");
 	printf("nk <keyname> - add key\n");
 	printf("dk <keyname> - delete key (must be empty. recursion not supported yet)\n");
 	printf("ed <value>            - Edit value\n");
 	printf("nv <type> <valuename> - Add value\n");
 	printf("dv <valuename>        - Delete value\n");
 	printf("delallv               - Delete all values in current key\n");
+	printf("rdel <keyname>        - Recursively delete key & subkeys\n");
 	printf("debug - enter buffer hexeditor\nq - quit\n");
         break;
 
@@ -1027,6 +1053,20 @@ void mainloop(void)
 	skipspace(&bp);
         nk_ls(hdesc, bp, cdofs+4, 0);
 	break;
+      case MCMD_CK :
+	bp++;
+	skipspace(&bp);
+        kv = get_class(hdesc, cdofs+4, bp);
+	if (kv) {
+	  hexdump((char *)&kv->data, 0, kv->len, 1);
+	  FREE(kv);
+	}
+	break;
+      case MCMD_RDEL :
+	bp++;
+	skipspace(&bp);
+        rdel_keys(hdesc, bp, cdofs+4);
+	break;
       case MCMD_EDIT :
 	bp++;
 	skipspace(&bp);
@@ -1066,7 +1106,12 @@ void mainloop(void)
       case MCMD_CAT:
 	bp++;
 	skipspace(&bp);
-	cat_vk(hdesc,cdofs+4,bp);
+	cat_vk(hdesc,cdofs+4,bp,0);
+	break;
+      case MCMD_CATHEX:
+	bp++;
+	skipspace(&bp);
+	cat_vk(hdesc,cdofs+4,bp,1);
 	break;
       case MCMD_STRUCT:
 	bp++;
@@ -1104,6 +1149,7 @@ int list_users(int pageit)
   int count = 0, countri = 0;
   struct ex_data ex;
 
+  if (H_SAM < 0) return(1);
   nkofs = trav_path(hive[H_SAM], 0,"\\SAM\\Domains\\Account\\Users\\Names\\",0);
   if (!nkofs) {
     printf("Cannot find usernames in registry! (is this a SAM-hive?)\n");
@@ -1151,7 +1197,7 @@ void find_n_change(char *username)
   struct keyval *v;
   int rid = 0;
 
-  if (!username) return;
+  if ((H_SAM < 0) || (!username)) return;
   if (*username == '0' && *(username+1) == 'x') sscanf(username,"%i",&rid);
   
   if (!rid) { /* Look up username */
