@@ -1,11 +1,13 @@
 /*
  * ntreg.c - NT Registry Hive access library
  *
+ * 2004-aug: Deep indirect index support. NT351 support. Recursive delete.
+ *           Debugged a lot in allocation routines. Still no expansion.
+ * 2004-jan: Verbosity updates
  * 2003-jan: Allocation of new data, supports adding/deleting keys & stuff.
  *           Missing is expanding the file.
  * 2003-jan: Seems there may be garbage pages at end of file, not zero pages
  *           now stops enumerating at first non 'hbin' page.
- * 2004-jan: Verbosity updates
  * 
  * NOTE: The API is not frozen. It can and will change every release.
  *
@@ -41,7 +43,11 @@
 
 #include "ntreg.h"
 
-const char ntreg_version[] = "ntreg lib routines, v0.91 040105, (c) Petter N Hagen";
+/* Set to abort() and debug on more critical errors */
+#define DOCORE 0
+
+
+const char ntreg_version[] = "ntreg lib routines, v0.92 040818, (c) Petter N Hagen";
 
 char *val_types[REG_MAX+1] = {
   "REG_NONE", "REG_SZ", "REG_EXPAND_SZ", "REG_BINARY", "REG_DWORD",       /* 0 - 4 */
@@ -526,6 +532,31 @@ void parse_li(struct hive *hdesc, int vofs, int blen)
 
 }
 
+/* Parse the ri subindex-datablock
+ * (Used to list li/lf/lh's when ~>500keys)
+ * vofs = offset into struct (after size linkage)
+ */
+void parse_ri(struct hive *hdesc, int vofs, int blen)
+{
+  struct ri_key *key;
+  int i;
+
+  printf("== ri at offset %0x\n",vofs);
+
+#define D_OFFS(o) ( (void *)&(key->o)-(void *)hdesc->buffer-vofs )
+
+  key = (struct ri_key *)(hdesc->buffer + vofs);
+  printf("%04x   number of subindices = %d\n", D_OFFS(no_lis), key->no_lis  );
+
+  for(i = 0; i < key->no_lis; i++) {
+    printf("%04x      %3d   Offset: 0x%0lx\n", 
+	   D_OFFS(hash[i].ofs_li), i,
+	   key->hash[i].ofs_li);
+  }
+  printf("== End of key info.\n");
+
+}
+
 
 /* Parse the datablock
  * vofs = offset into struct (after size linkage)
@@ -593,6 +624,9 @@ int parse_block(struct hive *hdesc, int vofs,int verbose)
     case 0x736b: /* sk */
       parse_sk(hdesc, vofs, seglen);
       break;
+    case 0x7269: /* ri */
+      parse_ri(hdesc, vofs, seglen);
+      break;
     default:
       printf("value data, or not handeled yet!\n");
       break;
@@ -648,7 +682,7 @@ int find_free_blk(struct hive *hdesc, int pofs, int size)
   
   p = (struct hbin_page *)(hdesc->buffer + pofs);
 
-  while (vofs-pofs < p->ofs_next) {
+  while (vofs-pofs < (p->ofs_next - HBIN_ENDFILL)) {
 
     seglen = get_int(hdesc->buffer+vofs);  
 
@@ -677,6 +711,13 @@ int find_free_blk(struct hive *hdesc, int pofs, int size)
 	if (seglen >= size) {
 #if FB_DEBUG
 	  printf("find_free_blk: found size %d block at 0x%x\n",seglen,vofs);
+#endif
+#if 0
+	  if (vofs == 0x19fb8) {
+	    printf("find_free_blk: vofs = %x, seglen = %x\n",vofs,seglen);
+	    debugit(hdesc->buffer,hdesc->size);
+	    abort();
+	  }
 #endif
 	  return(vofs);
 	}
@@ -842,6 +883,11 @@ int free_block(struct hive *hdesc, int blk)
   size = get_int(hdesc->buffer+blk);
   if (size >= 0) {
     printf("free_block: trying to free already free block!\n");
+#ifdef DOCORE
+      printf("blk = %x\n",blk);
+      debugit(hdesc->buffer,hdesc->size);
+      abort();
+#endif
     return(0);
   }
   size = -size;
@@ -853,35 +899,47 @@ int free_block(struct hive *hdesc, int blk)
   p = (struct hbin_page *)(hdesc->buffer + pofs);
   vofs = pofs + 0x20;
 
-  while (vofs-pofs < p->ofs_next) {
+  prevsz = -32;
 
-    seglen = get_int(hdesc->buffer+vofs);  
+  if (vofs != blk) {  /* Block is not at start of page? */
+    while (vofs-pofs < (p->ofs_next - HBIN_ENDFILL) ) {
 
-    if (seglen == 0) {
-      printf("free_block: EEEK! Zero data block size! (not registry or corrupt file?)\n");
+      seglen = get_int(hdesc->buffer+vofs);  
+      
+      if (seglen == 0) {
+	printf("free_block: EEEK! Zero data block size! (not registry or corrupt file?)\n");
+	debugit(hdesc->buffer,hdesc->size);
+	return(0);
+      }
+      
+      if (seglen < 0) {
+	seglen = -seglen;
+	/*      hexdump(hdesc->buffer,vofs,vofs+seglen+4,1); */
+      } 
+      prev = vofs;
+      vofs += seglen;
+      if (vofs == blk) break;
+    }
+    
+    if (vofs != blk) {
+      printf("free_block: ran off end of page!?!? Error in chains?\n");
+#ifdef DOCORE
+      printf("vofs = %x, pofs = %x, blk = %x\n",vofs,pofs,blk);
       debugit(hdesc->buffer,hdesc->size);
+      abort();
+#endif
       return(0);
     }
     
-    if (seglen < 0) {
-      seglen = -seglen;
-	/*      hexdump(hdesc->buffer,vofs,vofs+seglen+4,1); */
-    } 
-    prev = vofs;
-    vofs += seglen;
-    if (vofs == blk) break;
-  }
-
-  if (vofs != blk) {
-    printf("free_block: ran off end of page!?!? Error in chains?\n");
-    return(0);
+    prevsz = get_int(hdesc->buffer+prev);
+    
   }
 
   /* We also need details on next block (unless at end of page) */
   next = blk + size;
 
-  prevsz = get_int(hdesc->buffer+prev);
-  if (next >= p->ofs_next) nextsz = get_int(hdesc->buffer+next);
+  nextsz = 0;
+  if (next-pofs < (p->ofs_next - HBIN_ENDFILL) ) nextsz = get_int(hdesc->buffer+next);
 
 #if 0
   printf("offset prev : %x , blk: %x , next: %x\n",prev,blk,next);
@@ -889,7 +947,10 @@ int free_block(struct hive *hdesc, int blk)
 #endif
 
   /* Now check if next block is free, if so merge it with the one to be freed */
-  if ( (next >= p->ofs_next) && nextsz > 0) {
+  if ( nextsz > 0) {
+#if 0
+    printf("Swallow next\n");
+#endif
     size += nextsz;   /* Swallow it in current block */
     hdesc->useblk--;
     hdesc->usetot -= 4;
@@ -897,7 +958,7 @@ int free_block(struct hive *hdesc, int blk)
   }
 
   /* Now free the block (possibly with ajusted size as above) */
-  /*  bzero( (void *)(hdesc->buffer+blk), size); */
+   bzero( (void *)(hdesc->buffer+blk), size);
   *(int *)((hdesc->buffer)+blk) = (int)size;
   hdesc->usetot -= size;
   hdesc->unusetot -= size;
@@ -907,11 +968,14 @@ int free_block(struct hive *hdesc, int blk)
  
   /* Check if previous block is also free, if so, merge.. */
   if (prevsz > 0) {
+#if 0
+    printf("Swallow prev\n");
+#endif
     hdesc->usetot -= prevsz;
     hdesc->unusetot += prevsz;
     prevsz += size;
     /* And swallow current.. */
-    /*   bzero( (void *)(hdesc->buffer+prev), prevsz); */
+      bzero( (void *)(hdesc->buffer+prev), prevsz);
     *(int *)((hdesc->buffer)+prev) = (int)prevsz;
     hdesc->useblk--;
     return(prevsz);
@@ -921,6 +985,8 @@ int free_block(struct hive *hdesc, int blk)
 
 
 
+
+
 /* ================================================================ */
 
 /* ** Registry manipulation routines ** */
@@ -951,7 +1017,7 @@ int ex_next_n(struct hive *hdesc, int nkofs, int *count, int *countri, struct ex
     return(-1);
   }
 
-#undef EXNDEBUG
+#define EXNDEBUG 0
 
   lfkey = (struct lf_key *)(hdesc->buffer + key->ofs_lf + 0x1004);
   rikey = (struct ri_key *)(hdesc->buffer + key->ofs_lf + 0x1004);
@@ -963,10 +1029,16 @@ int ex_next_n(struct hive *hdesc, int nkofs, int *count, int *countri, struct ex
     if (*countri < 0 || *countri >= rikey->no_lis) { /* End of ri's? */
       return(0);
     }
-    /* Get the li-struct that's current based on countri */
+    /* Get the li of lf-struct that's current based on countri */
     likey = (struct li_key *)( hdesc->buffer + rikey->hash[*countri].ofs_li + 0x1004 ) ;
-    newnkofs = likey->hash[*count].ofs_nk + 0x1000;
-    /* Check if current li is exhausted */
+    if (likey->id == 0x696c) {
+      newnkofs = likey->hash[*count].ofs_nk + 0x1000;
+    } else {
+      lfkey = (struct lf_key *)( hdesc->buffer + rikey->hash[*countri].ofs_li + 0x1004 ) ;
+      newnkofs = lfkey->hash[*count].ofs_nk + 0x1000;
+    }
+
+    /* Check if current li/lf is exhausted */
 #if EXNDEBUG
     printf("likey->no_keys = %d\n",likey->no_keys);
 #endif
@@ -1176,6 +1248,7 @@ int trav_path(struct hive *hdesc, int vofs, char *path, int type)
   char part[ABSPATHLEN+1];
   char *partptr;
 
+  if (!hdesc) return(0);
   buf = hdesc->buffer;
 
   if (*path == '\\' && *(path+1) != '\\') {      /* Start from root if path starts with \ */
@@ -1237,6 +1310,10 @@ int trav_path(struct hive *hdesc, int vofs, char *path, int type)
       r = 0;
       likey = (struct li_key *)( hdesc->buffer + rikey->hash[r].ofs_li + 0x1004 ) ;
       subs = likey->no_keys;
+      if (likey->id != 0x696c) {  /* Bwah, not li anyway, XP uses lh usually which is actually smarter */
+	lfkey = (struct lf_key *)( hdesc->buffer + rikey->hash[r].ofs_li + 0x1004 ) ;
+	likey = NULL;
+      }
     } else {
       if (lfkey->id == 0x696c) { /* li? */
 	likey = (struct li_key *)(buf + lfofs);
@@ -1269,6 +1346,10 @@ int trav_path(struct hive *hdesc, int vofs, char *path, int type)
 	newnkofs = rikey->hash[r].ofs_li;
 	likey = (struct li_key *)( hdesc->buffer + newnkofs + 0x1004 ) ;
 	subs = likey->no_keys;
+	if (likey->id != 0x696c) {  /* Bwah, not li anyway, XP uses lh usually which is actually smarter */
+	  lfkey = (struct lf_key *)( hdesc->buffer + rikey->hash[r].ofs_li + 0x1004 ) ;
+	  likey = NULL;
+	}
       }
     } while (r < ricnt && ricnt);
 
@@ -1277,6 +1358,7 @@ int trav_path(struct hive *hdesc, int vofs, char *path, int type)
   return(0);
 }
 
+
 /* ls - list a 'nk' nodes subkeys and values
  * vofs - offset to start of data (skipping block linkage)
  * type - 0 = full, 1 = keys only. 2 = values only
@@ -1294,7 +1376,6 @@ void nk_ls(struct hive *hdesc, char *path, int vofs, int type)
 
   if(!nkofs) {
     printf("nk_ls: Key <%s> not found\n",path);
-    abort();
     return;
   }
   nkofs += 4;
@@ -1309,12 +1390,14 @@ void nk_ls(struct hive *hdesc, char *path, int vofs, int type)
     
   }
   
-  printf("Node has %ld subkeys and %ld values\n",key->no_subkeys,key->no_values);
+  printf("Node has %ld subkeys and %ld values",key->no_subkeys,key->no_values);
+  if (key->len_classnam) printf(", and class-data of %d bytes",key->len_classnam);
+  printf("\n");
 
   if (key->no_subkeys) {
     printf("offs          key name\n");
     while ((ex_next_n(hdesc, nkofs, &count, &countri, &ex) > 0)) {
-      printf("[%6x]   <%s>\n", ex.nkoffs, ex.name);
+      printf("[%6x]   %c  <%s>\n", ex.nkoffs, (ex.nk->len_classnam)?'*':' ',ex.name);
       FREE(ex.name);
     }
   }
@@ -1460,6 +1543,33 @@ int get_dword(struct hive *hdesc, int vofs, char *path)
   
 }
 
+/* Sanity checker when transferring data into a block
+ * ofs = offset to data block, point to start of actual datablock linkage
+ * data = data to copy
+ * size = size of data to copy
+ */
+
+int fill_block(struct hive *hdesc, int ofs, void *data, int size)
+{
+  int blksize;
+
+  blksize = get_int(hdesc->buffer + ofs);
+  blksize = -blksize;
+
+#if 0
+  printf("fill_block: ofs = %x - %x, size = %x, blksize = %x\n",ofs,ofs+size,size,blksize);
+#endif
+  /*  if (blksize < size || ( (ofs & 0xfffff000) != ((ofs+size) & 0xfffff000) )) { */
+  if (blksize < size) {
+    printf("fill_block: ERROR: block to small for data: ofs = %x, size = %x, blksize = %x\n",ofs,size,blksize);
+    debugit(hdesc->buffer,hdesc->size);
+    abort();
+  }
+
+  memcpy(hdesc->buffer + ofs + 4, data, size);
+  return(0);
+}
+
 
 /* Free actual data of a value, and update value descriptor
  * hdesc - hive
@@ -1654,7 +1764,7 @@ void del_allvalues(struct hive *hdesc, int nkofs)
   }
 
   if (!nk->no_values) {
-    printf("del_avalues: Key has no values!\n");
+    /*    printf("del_avalues: Key has no values!\n"); */
     return;
   }
 
@@ -1744,6 +1854,9 @@ int del_value(struct hive *hdesc, int nkofs, char *name)
   return(0);
 }
 
+
+
+
 /* Add a subkey to a key
  * hdesc - usual..
  * nkofs - offset of current nk
@@ -1751,11 +1864,17 @@ int del_value(struct hive *hdesc, int nkofs, char *name)
  * return: ptr to new keystruct, or NULL
  */
 
+#undef AKDEBUG
 struct nk_key *add_key(struct hive *hdesc, int nkofs, char *name)
 {
 
-  int slot = 0, newlfofs = 0, oldlfofs = 0, o, n, i, onkofs, newnkofs, cmp;
-  struct lf_key *newlf, *oldlf;
+  int slot, newlfofs = 0, oldlfofs = 0, newliofs = 0;
+  int oldliofs = 0;
+  int o, n, i, onkofs, newnkofs, cmp;
+  int rimax, rislot, riofs, namlen;
+  struct ri_key *ri = NULL;
+  struct lf_key *newlf = NULL, *oldlf;
+  struct li_key *newli = NULL, *oldli;
   struct nk_key *key, *newnk, *onk;
   long hash;
 
@@ -1766,54 +1885,159 @@ struct nk_key *add_key(struct hive *hdesc, int nkofs, char *name)
     return(NULL);
   }
 
+  namlen = strlen(name);
+
   slot = -1;
   if (key->no_subkeys) {   /* It already has subkeys */
     
     oldlfofs = key->ofs_lf;
+    oldliofs = key->ofs_lf;
+   
     oldlf = (struct lf_key *)(hdesc->buffer + oldlfofs + 0x1004);
-    if (oldlf->id != 0x666c && oldlf->id != 0x686c) {
-      printf("add_key: index other than 'lf' or 'lh' not supported yet. 0x%04x\n",oldlf->id);
+    if (oldlf->id != 0x666c && oldlf->id != 0x686c && oldlf->id != 0x696c && oldlf->id != 0x6972)  {
+      printf("add_key: index type not supported: 0x%04x\n",oldlf->id);
       return(NULL);
     }
 
-    ALLOC(newlf, 8 + 8*oldlf->no_keys + 8, 1);
-    newlf->no_keys = oldlf->no_keys + 1;
-    newlf->id = oldlf->id;
+    rimax = 0; ri = NULL; riofs = 0; rislot = -1;
+    if (oldlf->id == 0x6972) {  /* Indirect index 'ri', init loop */
+      riofs = key->ofs_lf;
+      ri = (struct ri_key *)(hdesc->buffer + riofs + 0x1004);
+      rimax = ri->no_lis-1;
+
+#ifdef AKDEBUG
+      printf("add_key: entering 'ri' traverse, rimax = %d\n",rimax);
+#endif
+
+      oldliofs = ri->hash[rislot+1].ofs_li;
+      oldlfofs = ri->hash[rislot+1].ofs_li;
 
-    /* Now copy old, checking where to insert (alphabetically) */
-    for (o = 0, n = 0; o < oldlf->no_keys; o++,n++) {
-      onkofs = oldlf->hash[o].ofs_nk;
-      onk = (struct nk_key *)(onkofs + hdesc->buffer + 0x1004);
-      if (slot == -1) {
-	cmp = strncasecmp(name, onk->keyname, onk->len_name);
-	if (!cmp) {
-	  printf("add_key: key %s already exists!\n",name);
-	  FREE(newlf);
-	  return(NULL);
-	}
-	if ( cmp < 0) {
-	  slot = o;
-	  n++;
-	}
-      }
-      newlf->hash[n].ofs_nk = oldlf->hash[o].ofs_nk;
-      newlf->hash[n].name[0] = oldlf->hash[o].name[0];
-      newlf->hash[n].name[1] = oldlf->hash[o].name[1];
-      newlf->hash[n].name[2] = oldlf->hash[o].name[2];
-      newlf->hash[n].name[3] = oldlf->hash[o].name[3];
     }
-    if (slot == -1) slot = oldlf->no_keys;
-  } else {
+
+    do {   /* 'ri' loop, at least run once if no 'ri' deep index */
+
+      if (ri) { /* Do next 'ri' slot */
+	rislot++;
+	oldliofs = ri->hash[rislot].ofs_li;
+	oldlfofs = ri->hash[rislot].ofs_li;
+	oldli = (struct li_key *)(hdesc->buffer + oldliofs + 0x1004);
+	oldlf = (struct lf_key *)(hdesc->buffer + oldlfofs + 0x1004);
+      }
+
+      oldli = (struct li_key *)(hdesc->buffer + oldliofs + 0x1004);
+      oldlf = (struct lf_key *)(hdesc->buffer + oldlfofs + 0x1004);
+
+#ifdef AKDEBUG
+      printf("add_key: top of ri-loop: rislot = %d\n",rislot);
+#endif
+      slot = -1;
+
+      if (oldli->id == 0x696c) {  /* li */
+	
+#ifdef AKDEBUG
+	printf("add_key: li slot allocate\n");
+#endif	
+
+	FREE(newli);
+	ALLOC(newli, 8 + 4*oldli->no_keys + 4, 1);
+	newli->no_keys = oldli->no_keys;
+	newli->id = oldli->id;
+	
+	/* Now copy old, checking where to insert (alphabetically) */
+	for (o = 0, n = 0; o < oldli->no_keys; o++,n++) {
+	  onkofs = oldli->hash[o].ofs_nk;
+	  onk = (struct nk_key *)(onkofs + hdesc->buffer + 0x1004);
+	  if (slot == -1) {
+#if 0
+	    printf("add_key: cmp <%s> with <%s>\n",name,onk->keyname);
+#endif
+
+	    cmp = strncasecmp(name, onk->keyname, (namlen > onk->len_name) ? namlen : onk->len_name);
+	    if (!cmp) {
+	      printf("add_key: key %s already exists!\n",name);
+	      FREE(newli);
+	      return(NULL);
+	    }
+	    if ( cmp < 0) {
+	      slot = o;
+	      rimax = rislot; /* Cause end of 'ri' search, too */
+	      n++;
+#ifdef AKDEBUG
+	      printf("add_key: li-match: slot = %d\n",o);
+#endif
+	    }
+	  }
+	  newli->hash[n].ofs_nk = oldli->hash[o].ofs_nk;
+	}
+	if (slot == -1) slot = oldli->no_keys;
+	
+      } else { /* lf or lh */
+
+	oldlf = (struct lf_key *)(hdesc->buffer + oldlfofs + 0x1004);
+	
+	FREE(newlf);
+	ALLOC(newlf, 8 + 8*oldlf->no_keys + 8, 1);
+	newlf->no_keys = oldlf->no_keys;
+	newlf->id = oldlf->id;
+#ifdef AKDEBUG	
+	printf("add_key: new lf/lh no_keys: %d\n",newlf->no_keys);
+#endif
+	
+	/* Now copy old, checking where to insert (alphabetically) */
+	for (o = 0, n = 0; o < oldlf->no_keys; o++,n++) {
+	  onkofs = oldlf->hash[o].ofs_nk;
+	  onk = (struct nk_key *)(onkofs + hdesc->buffer + 0x1004);
+	  if (slot == -1) {
+
+#if 0
+	    printf("add_key: cmp <%s> with <%s>\n",name,onk->keyname);
+#endif
+	    cmp = strncasecmp(name, onk->keyname, (namlen > onk->len_name) ? namlen : onk->len_name);
+	    if (!cmp) {
+	      printf("add_key: key %s already exists!\n",name);
+	      FREE(newlf);
+	      return(NULL);
+	    }
+	    if ( cmp < 0) {
+	      slot = o;
+	      rimax = rislot;  /* Cause end of 'ri' search, too */
+	      n++;
+#ifdef AKDEBUG
+	      printf("add_key: lf-match: slot = %d\n",o);
+#endif
+	    }
+	  }
+	  newlf->hash[n].ofs_nk = oldlf->hash[o].ofs_nk;
+	  newlf->hash[n].name[0] = oldlf->hash[o].name[0];
+	  newlf->hash[n].name[1] = oldlf->hash[o].name[1];
+	  newlf->hash[n].name[2] = oldlf->hash[o].name[2];
+	  newlf->hash[n].name[3] = oldlf->hash[o].name[3];
+	}
+	if (slot == -1) slot = oldlf->no_keys;
+      } /* li else check */
+
+
+
+    } while ( (rislot < rimax) );  /* 'ri' wrapper loop */
+
+  } else { /* Parent was empty, make new index block */
+#ifdef AKDEBUG
+    printf("add_key: new index!\n");
+#endif
     ALLOC(newlf, 8 + 8, 1);
     newlf->no_keys = 1;
-    newlf->id = 0x666c;  
-  }
+    /* Use ID (lf, lh or li) we fetched from root node, so we use same as rest of hive */
+    newlf->id = hdesc->nkindextype;
+    slot = 0;
+  } /* if has keys before */
+
 
   /* Make and fill in new nk */
   newnkofs = alloc_block(hdesc, nkofs, sizeof(struct nk_key) + strlen(name));
   if (!newnkofs) {
     printf("add_key: unable to allocate space for new key descriptor for %s!\n",name);
     FREE(newlf);
+    FREE(newli);
     return(NULL);
   }
   newnk = (struct nk_key *)(hdesc->buffer + newnkofs + 4);
@@ -1829,40 +2053,79 @@ struct nk_key *add_key(struct hive *hdesc, int nkofs, char *name)
   newnk->ofs_classnam  = -1;
   newnk->len_name      = strlen(name);
   newnk->len_classnam  = 0;
-  strncpy(newnk->keyname, name, strlen(name));
+  strcpy(newnk->keyname, name);
   
-  /* And put its offset into parents index list */
-  newlf->hash[slot].ofs_nk = newnkofs - 0x1000;
-  if (newlf->id == 0x666c) {        /* lf hash */
-    newlf->hash[slot].name[0] = 0;
-    newlf->hash[slot].name[1] = 0;
-    newlf->hash[slot].name[2] = 0;
-    newlf->hash[slot].name[3] = 0;
-    strncpy(newlf->hash[n].name, name, 4);
-  } else if (newlf->id == 0x686c) {  /* lh. XP uses this. hashes whole name */
+  if (newli) {  /* Handle li */
+
+#if AKDEBUG
+    printf("add_key: li fill at slot: %d\n",slot);
+#endif
+
+    /* And put its offset into parents index list */
+    newli->hash[slot].ofs_nk = newnkofs - 0x1000;
+    newli->no_keys++;
+    
+    /* Allocate space for our new li list and copy it into reg */
+    newliofs = alloc_block(hdesc, nkofs, 8 + 4*newli->no_keys);
+    if (!newliofs) {
+      printf("add_key: unable to allocate space for new index table for %s!\n",name);
+      FREE(newli);
+      free_block(hdesc,newnkofs);
+      return(NULL);
+    }
+    /*    memcpy(hdesc->buffer + newliofs + 4, newli, 8 + 4*newli->no_keys); */
+    fill_block(hdesc, newliofs, newli, 8 + 4*newli->no_keys);
+
+
+  } else {  /* lh or lf */
+
+#ifdef AKDEBUG
+    printf("add_key: lf/lh fill at slot: %d, rislot: %d\n",slot,rislot);
+#endif
+    /* And put its offset into parents index list */
+    newlf->hash[slot].ofs_nk = newnkofs - 0x1000;
+    newlf->no_keys++;
+    if (newlf->id == 0x666c) {        /* lf hash */
+      newlf->hash[slot].name[0] = 0;
+      newlf->hash[slot].name[1] = 0;
+      newlf->hash[slot].name[2] = 0;
+      newlf->hash[slot].name[3] = 0;
+      strncpy(newlf->hash[slot].name, name, 4);
+    } else if (newlf->id == 0x686c) {  /* lh. XP uses this. hashes whole name */
       for (i = 0,hash = 0; i < strlen(name); i++) {
 	hash *= 37;
 	hash += toupper(name[i]);
       }
       newlf->lh_hash[slot].hash = hash;
-  }
+    }
+    
+    /* Allocate space for our new lf list and copy it into reg */
+    newlfofs = alloc_block(hdesc, nkofs, 8 + 8*newlf->no_keys);
+    if (!newlfofs) {
+      printf("add_key: unable to allocate space for new index table for %s!\n",name);
+      FREE(newlf);
+      free_block(hdesc,newnkofs);
+      return(NULL);
+    }
+    /*    memcpy(hdesc->buffer + newlfofs + 4, newlf, 8 + 8*newlf->no_keys); */
+    fill_block(hdesc, newlfofs, newlf, 8 + 8*newlf->no_keys);
+    
+  } /* li else */
 
-  /* Allocate space for our new lf list and copy it into reg */
-  newlfofs = alloc_block(hdesc, nkofs, 8 + 8*newlf->no_keys);
-  if (!newlfofs) {
-    printf("add_key: unable to allocate space for new key descriptor for %s!\n",name);
-    FREE(newlf);
-    free_block(hdesc,newnkofs);
-    return(NULL);
-  }
-  memcpy(hdesc->buffer + newlfofs + 4, newlf, 8 + 8*newlf->no_keys);
 
   /* Update parent, and free old lf list */
   key->no_subkeys++;
-  key->ofs_lf = newlfofs - 0x1000;
-  if (oldlfofs) free_block(hdesc,oldlfofs + 0x1000);
+  if (ri) {  /* ri index */
+    ri->hash[rislot].ofs_li = (newlf ? newlfofs : newliofs) - 0x1000;
+  } else { /* Parent key */
+    key->ofs_lf = (newlf ? newlfofs : newliofs) - 0x1000;
+  }
+
+  if (newlf && oldlfofs) free_block(hdesc,oldlfofs + 0x1000);
+  if (newli && oldliofs) free_block(hdesc,oldliofs + 0x1000);
 
   FREE(newlf);
+  FREE(newli);
   return(newnk);
 
 
@@ -1871,19 +2134,29 @@ struct nk_key *add_key(struct hive *hdesc, int nkofs, char *name)
 /* Delete a subkey from a key
  * hdesc - usual..
  * nkofs - offset of current nk
- * name  - name of key to delete
+ * name  - name of key to delete (must match exactly, also case)
  * return: 1 - err, 0 - ok
  */
 
+#undef DKDEBUG
+
 int del_key(struct hive *hdesc, int nkofs, char *name)
 {
 
   int slot = 0, newlfofs = 0, oldlfofs = 0, o, n, onkofs,  delnkofs;
-  struct lf_key *newlf, *oldlf;
+  int oldliofs = 0, no_keys = 0, newriofs = 0;
+  int namlen;
+  int rimax, riofs, rislot;
+  struct ri_key *ri, *newri = NULL;
+  struct lf_key *newlf = NULL, *oldlf = NULL;
+  struct li_key *newli = NULL, *oldli = NULL;
   struct nk_key *key, *onk, *delnk;
+  char fullpath[501];
 
   key = (struct nk_key *)(hdesc->buffer + nkofs);
 
+  namlen = strlen(name);
+
   if (key->id != 0x6b6e) {
     printf("add_key: current ptr not nk\n");
     return(1);
@@ -1896,68 +2169,305 @@ int del_key(struct hive *hdesc, int nkofs, char *name)
   }
 
   oldlfofs = key->ofs_lf;
+  oldliofs = key->ofs_lf;
+  
   oldlf = (struct lf_key *)(hdesc->buffer + oldlfofs + 0x1004);
-  if (oldlf->id != 0x666c && oldlf->id != 0x686c) {
-    printf("del_key: index other than 'lf' or 'lh' not supported yet. 0x%04x\n",oldlf->id);
+  if (oldlf->id != 0x666c && oldlf->id != 0x686c && oldlf->id != 0x696c && oldlf->id != 0x6972)  {
+    printf("del_key: index other than 'lf', 'li' or 'lh' not supported yet. 0x%04x\n",oldlf->id);
     return(1);
   }
 
-  ALLOC(newlf, 8 + 8*oldlf->no_keys - 8, 1);
-  newlf->no_keys = oldlf->no_keys - 1;
-  newlf->id = oldlf->id;
+  rimax = 0; ri = NULL; riofs = 0; rislot = 0;
 
-  /* Now copy old, checking where to delete */
-  for (o = 0, n = 0; o < oldlf->no_keys; o++,n++) {
-    onkofs = oldlf->hash[o].ofs_nk;
-    onk = (struct nk_key *)(onkofs + hdesc->buffer + 0x1004);
-    if (slot == -1 && !strncmp(name, onk->keyname, onk->len_name)) {
-        slot = o;
-	delnkofs = onkofs; delnk = onk;
-	o++;
-    }
-    newlf->hash[n].ofs_nk = oldlf->hash[o].ofs_nk;
-    newlf->hash[n].name[0] = oldlf->hash[o].name[0];
-    newlf->hash[n].name[1] = oldlf->hash[o].name[1];
-    newlf->hash[n].name[2] = oldlf->hash[o].name[2];
-    newlf->hash[n].name[3] = oldlf->hash[o].name[3];
+  if (oldlf->id == 0x6972) {  /* Indirect index 'ri', init loop */
+    riofs = key->ofs_lf;
+    ri = (struct ri_key *)(hdesc->buffer + riofs + 0x1004);
+    rimax = ri->no_lis-1;
+    
+#ifdef DKDEBUG
+    printf("del_key: entering 'ri' traverse, rimax = %d\n",rimax);
+#endif
+    
+    oldliofs = ri->hash[rislot+1].ofs_li;
+    oldlfofs = ri->hash[rislot+1].ofs_li;
+    
   }
+  
+  do {   /* 'ri' loop, at least run once if no 'ri' deep index */
+    
+    if (ri) { /* Do next 'ri' slot */
+      rislot++;
+      oldliofs = ri->hash[rislot].ofs_li;
+      oldlfofs = ri->hash[rislot].ofs_li;
+      oldli = (struct li_key *)(hdesc->buffer + oldliofs + 0x1004);
+      oldlf = (struct lf_key *)(hdesc->buffer + oldlfofs + 0x1004);
+    }
+    
+    oldli = (struct li_key *)(hdesc->buffer + oldliofs + 0x1004);
+    oldlf = (struct lf_key *)(hdesc->buffer + oldlfofs + 0x1004);
+    
+#ifdef DKDEBUG
+    printf("del_key: top of ri-loop: rislot = %d\n",rislot);
+#endif
+    slot = -1;
+    
+    if (oldlf->id == 0x696c) {   /* 'li' handler */
+#ifdef DKDEBUG      
+      printf("del_key: li handler\n");
+#endif
+      
+      FREE(newli);
+      ALLOC(newli, 8 + 4*oldli->no_keys - 4, 1);
+      newli->no_keys = oldli->no_keys - 1; no_keys = newli->no_keys;
+      newli->id = oldli->id;
+      
+      /* Now copy old, checking where to delete */
+      for (o = 0, n = 0; o < oldli->no_keys; o++,n++) {
+	onkofs = oldli->hash[o].ofs_nk;
+	onk = (struct nk_key *)(onkofs + hdesc->buffer + 0x1004);
+	if (slot == -1 && onk->len_name == namlen && !strncmp(name, onk->keyname, (onk->len_name > namlen) ? onk->len_name : namlen)) {
+	  slot = o;
+	  delnkofs = onkofs; delnk = onk;
+	  rimax = rislot;
+	  o++;
+	}
+	newli->hash[n].ofs_nk = oldli->hash[o].ofs_nk;
+      }
+      
+      
+    } else { /* 'lf' or 'lh' are similar */
+      
+#ifdef DKDEBUG
+      printf("del_key: lf or lh handler\n");
+#endif
+      FREE(newlf);
+      ALLOC(newlf, 8 + 8*oldlf->no_keys - 8, 1);
+      newlf->no_keys = oldlf->no_keys - 1; no_keys = newlf->no_keys;
+      newlf->id = oldlf->id;
+      
+      /* Now copy old, checking where to delete */
+      for (o = 0, n = 0; o < oldlf->no_keys; o++,n++) {
+	onkofs = oldlf->hash[o].ofs_nk;
+	onk = (struct nk_key *)(onkofs + hdesc->buffer + 0x1004);
+	if (slot == -1 && (onk->len_name == namlen) && !strncmp(name, onk->keyname, onk->len_name)) {
+	  slot = o;
+	  delnkofs = onkofs; delnk = onk;
+	  rimax = rislot;
+	  o++;
+	}
+	newlf->hash[n].ofs_nk = oldlf->hash[o].ofs_nk;
+	newlf->hash[n].name[0] = oldlf->hash[o].name[0];
+	newlf->hash[n].name[1] = oldlf->hash[o].name[1];
+	newlf->hash[n].name[2] = oldlf->hash[o].name[2];
+	newlf->hash[n].name[3] = oldlf->hash[o].name[3];
+      }
+    } /* else lh or lf */
+
+  } while (rislot < rimax);  /* ri traverse loop */
+
   if (slot == -1) {
     printf("del_key: subkey %s not found!\n",name);
     FREE(newlf);
+    FREE(newli);
     return(1);
   }
 
+#ifdef DKDEBUG
+  printf("del_key: key found at slot %d\n",slot);
+#endif
+
   if (delnk->no_values || delnk->no_subkeys) {
     printf("del_key: subkey %s has subkeys or values. Not deleted.\n",name);
     FREE(newlf);
+    FREE(newli);
     return(1);
   }
 
   /* Allocate space for our new lf list and copy it into reg */
-  if (newlf->no_keys) {
-    newlfofs = alloc_block(hdesc, nkofs, 8 + 8*newlf->no_keys);
+  if ( no_keys && (newlf || newli) ) {
+    newlfofs = alloc_block(hdesc, nkofs, 8 + (newlf ? 8 : 4) * no_keys);
+#ifdef DKDEBUG
+    printf("del_key: alloc_block for index returns: %x\n",newlfofs);
+#endif
     if (!newlfofs) {
-      printf("del_key: unable to allocate space for new key descriptor for %s! Not deleted\n",name);
+      printf("del_key: WARNING: unable to allocate space for new key descriptor for %s! Not deleted\n",name);
       FREE(newlf);
       return(1);
     }
-    memcpy(hdesc->buffer + newlfofs + 4, newlf, 8 + 8*newlf->no_keys);
+
+    /*    memcpy(hdesc->buffer + newlfofs + 4,
+	   ((void *)newlf ? (void *)newlf : (void *)newli), 8 + (newlf ? 8 : 4) * no_keys);
+    */
+    fill_block(hdesc, newlfofs,
+	   ((void *)newlf ? (void *)newlf : (void *)newli), 8 + (newlf ? 8 : 4) * no_keys);
+
+
   } else {  /* Last deleted, will throw away index */
-    newlfofs = -1;
+    newlfofs = 0xfff;  /* We subtract 0x1000 later */
   }
 
+  if (newlfofs < 0xfff) {
+    printf("del_key: ERROR: newlfofs = %x\n",newlfofs);
+    debugit(hdesc->buffer,hdesc->size);
+    abort();
+  }
+
+  /* Check for CLASS data, if so, deallocate it too */
+  if (delnk->len_classnam) {
+    free_block(hdesc, delnk->ofs_classnam + 0x1000);
+  }
   /* Now it's safe to zap the nk */
   free_block(hdesc, delnkofs + 0x1000);
+  /* And the old index list */
+  free_block(hdesc, (oldlfofs ? oldlfofs : oldliofs) + 0x1000);
 
   /* Update parent */
   key->no_subkeys--;
-  key->ofs_lf = newlfofs - 0x1000;
+
+  if (ri) {
+    if (newlfofs == 0xfff) {
+
+      *fullpath = 0;
+      get_abs_path(hdesc, nkofs, fullpath, 480);
+
+      printf("del_key: need to delete ri-entry! %x - %s\n",nkofs,fullpath);
+
+      if (ri->no_lis > 1) {  /* We have subindiceblocks left? */
+	/* Delete from array */
+	ALLOC(newri, 8 + 4*ri->no_lis - 4, 1);
+	newri->no_lis = ri->no_lis - 1;
+	newri->id = ri->id;
+	for (o = 0, n = 0; o < ri->no_lis; o++,n++) {
+	  if (n == rislot) o++;
+	  newri->hash[n].ofs_li = ri->hash[o].ofs_li;
+	}
+	newriofs = alloc_block(hdesc, nkofs, 8 + newri->no_lis*4 );
+	printf("del_key: alloc_block for ri-block returns: %x\n",newriofs);
+	if (!newriofs) {
+	  printf("del_key: WARNING: unable to allocate space for ri-index for %s! Not deleted\n",name);
+	  FREE(newlf);
+	  FREE(newri);
+	  return(1);
+	}
+	fill_block(hdesc, newriofs, newri, 8 + newri->no_lis * 4);
+	free_block(hdesc, riofs + 0x1000);
+	key->ofs_lf = newriofs - 0x1000;
+	FREE(newri);
+      } else { /* Last entry in ri was deleted, get rid of it, key is empty */
+	printf("del_key: last ri deleted for %x\n",nkofs);
+	free_block(hdesc, riofs + 0x1000);
+	key->ofs_lf = -1;
+      }
+    } else {
+      ri->hash[rislot].ofs_li = newlfofs - 0x1000; 
+    }
+  } else {
+    key->ofs_lf = newlfofs - 0x1000;
+  }
 
   FREE(newlf);
   return(0);
 
 }
 
+/* Recursive delete keys
+ * hdesc - usual..
+ * nkofs - offset of current nk
+ * name  - name of key to delete
+ * return: 0 - ok, 1 fail
+ */
+void rdel_keys(struct hive *hdesc, char *path, int vofs)
+{
+  struct nk_key *key;
+  int nkofs;
+  struct ex_data ex;
+  int count = 0, countri = 0;
+  
+
+  if (!path || !*path) return;
+
+  nkofs = trav_path(hdesc, vofs, path, 0);
+
+  if(!nkofs) {
+    printf("rdel_keys: Key <%s> not found\n",path);
+    return;
+  }
+  nkofs += 4;
+
+  key = (struct nk_key *)(hdesc->buffer + nkofs);
+#if 0
+  printf("rdel of node at offset 0x%0x\n",nkofs);
+#endif
+  if (key->id != 0x6b6e) {
+    printf("Error: Not a 'nk' node!\n");
+
+    debugit(hdesc->buffer,hdesc->size);
+    
+  }
+  
+#if 0
+  printf("Node has %ld subkeys and %ld values\n",key->no_subkeys,key->no_values);
+#endif
+  if (key->no_subkeys) {
+    while ((ex_next_n(hdesc, nkofs, &count, &countri, &ex) > 0)) {
+#if 0
+      printf("%s\n",ex.name);
+#endif
+      rdel_keys(hdesc, ex.name, nkofs);
+      count = 0;
+      countri = 0;
+      FREE(ex.name);
+    }
+  }
+
+  del_allvalues(hdesc, nkofs);
+  del_key(hdesc, key->ofs_parent+0x1004, path);
+
+}
+  
+
+/* Get and copy keys CLASS-data (if any) to buffer
+ * Returns a buffer with the data (first long is size). see ntreg.h
+ * NOTE: caller must deallocate buffer! a simple free(keyval) will suffice.
+ */
+struct keyval *get_class(struct hive *hdesc,
+			    int curnk, char *path)
+{
+  int clen = 0, dofs = 0, nkofs;
+  struct nk_key *key;
+  struct keyval *data;
+  void *classdata;
+
+  if (!path && !curnk) return(NULL);
+
+  nkofs = trav_path(hdesc, curnk, path, 0);
+
+  if(!nkofs) {
+    printf("get_class: Key <%s> not found\n",path);
+    return(NULL);
+  }
+  nkofs += 4;
+  key = (struct nk_key *)(hdesc->buffer + nkofs);
+
+  clen = key->len_classnam;
+  if (!clen) {
+    printf("get_class: Key has no class data.\n");
+    return(NULL);
+  }
+
+  dofs = key->ofs_classnam;
+  classdata = (void *)(hdesc->buffer + dofs + 0x1004);
+  
+#if 0
+  printf("get_class: len_classnam = %d\n",clen);
+  printf("get_class: ofs_classnam = 0x%x\n",dofs);
+#endif
+
+  ALLOC(data, sizeof(struct keyval) + clen,1);
+  data->len = clen;
+  memcpy(&data->data, classdata, clen);
+  return(data);
+}
 
 
 /* Write to registry value.
@@ -2022,6 +2532,9 @@ void closeHive(struct hive *hdesc)
 {
 
   printf("closing hive %s\n",hdesc->filename);
+  if (hdesc->state & HMODE_OPEN) {
+    close(hdesc->filedesc);
+  }
   FREE(hdesc->filename);
   FREE(hdesc->buffer);
   FREE(hdesc);
@@ -2067,6 +2580,8 @@ struct hive *openHive(char *filename, int mode)
   char *c;
   struct hbin_page *p;
   struct regf_header *hdr;
+  struct nk_key *nk;
+  struct ri_key *rikey;
   int verbose = (mode & HMODE_VERBOSE);
 
   CREATE(hdesc,struct hive,1);
@@ -2130,7 +2645,35 @@ struct hive *openHive(char *filename, int mode)
    for (c = hdr->name; *c && (c < hdr->name + 64); c += 2) putchar(*c);
 
    hdesc->rootofs = hdr->ofs_rootkey + 0x1000;
-      printf(">\nROOT KEY at offset: 0x%06x\n",hdesc->rootofs);
+   printf(">\nROOT KEY at offset: 0x%06x * ",hdesc->rootofs);
+
+   /* Cache the roots subkey index type (li,lf,lh) so we can use the correct
+    * one when creating the first subkey in a key */
+   
+   nk = (struct nk_key *)(hdesc->buffer + hdesc->rootofs + 4);
+   if (nk->id == 0x6b6e) {
+     rikey = (struct ri_key *)(hdesc->buffer + nk->ofs_lf + 0x1004);
+     hdesc->nkindextype = rikey->id;
+     if (hdesc->nkindextype == 0x6972) {  /* Gee, big root, must check indirectly */
+       printf("load_hive: DEBUG: BIG ROOT!\n");
+       rikey = (struct ri_key *)(hdesc->buffer + rikey->hash[0].ofs_li + 0x1004);
+       hdesc->nkindextype = rikey->id;
+     }
+     if (hdesc->nkindextype != 0x666c &&
+	 hdesc->nkindextype != 0x686c &&
+	 hdesc->nkindextype != 0x696c) {
+       hdesc->nkindextype = 0x666c;
+     }
+
+     printf("Subkey indexing type is: %04x <%c%c>\n",
+	    hdesc->nkindextype,
+	    hdesc->nkindextype & 0xff,
+	    hdesc->nkindextype >> 8);
+   } else {
+     printf("load_hive: WARNING: ROOT key does not seem to be a key! (not type == nk)\n");
+   }
+
+
 
    while (pofs < hdesc->size) {
 #ifdef LOAD_DEBUG
